;; -*- Mode: Lisp; Package: VISU-SMOTS -*-

;;;  (c) copyright 2001 by
;;;           Irène Durand (idurand@labri.fr)

;;; This library is free software; you can redistribute it and/or
;;; modify it under the terms of the GNU Library General Public
;;; License as published by the Free Software Foundation; either
;;; version 2 of the License, or (at your option) any later version.
;;;
;;; This library is distributed in the hope that it will be useful,
;;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
;;; Library General Public License for more details.
;;;
;;; You should have received a copy of the GNU Library General Public
;;; License along with this library; if not, write to the
;;; Free Software Foundation, Inc., 59 Temple Place - Suite 330,
;;; Boston, MA  02111-1307  USA.

;(setf interface:*interface-style* :tty)
#+cmu
(setf ext:*gc-verbose* nil)

(in-package :visu-smots)
(defconstant +largeur-texte+ (1+ (length "----------")))
;;(defvar *progress-pane* nil)
;;(defvar *processname-pane*)
(defvar *data-directory-field*)

(defun normalize-string (s n)
  (let ((l (length s)))
    (if (>= l n)
        s
        (concatenate 'string " " s (make-string (- n l) :initial-element #\space)))))

(let*
    ((counter 0)
     (chars (make-array 4 :initial-contents '("\/" "\-" "\\" "\|")))
     (tog 0)
     (len (length chars)))
  (defun next-char ()
    (setf counter (mod (1+ counter) 100000))
    (if (zerop (mod counter 1000))
        (progn
          (setf tog (mod (1+ tog) len))
          (values t (aref chars tog)))
      (values nil (aref chars tog)))))

(defun interface ()
  (run-test 'visu-smots)
  0)

(defun start ()
  (init)
  (interface))

(defun start-darwin ()
  (init)
  (interface)
  (cl-user::quit))

(defun run-test (name)
  (loop for port in climi::*all-ports*
	do (destroy-port port))
  (setq climi::*all-ports* nil)
  (let ((frame (make-application-frame name)))
	(run-frame-top-level frame)))

(defvar *cpt* 0)

;; (defmethod display-step :around ()
;;   (when (and *progress-pane* (gadget-value *progress-pane*))
;;     (multiple-value-bind (res char)
;; 	(next-char)
;;       (when res
;; 	(setf (gadget-value *progress-pane*)
;; 	      (normalize-string char +largeur-texte+))))))

(defun pause ()
  (incf *cpt*)
  (dotimes (i 4000000000)
    (1+ 1))
  (format *output-stream* "f ~A done~%" *cpt*))

(make-command-table
 'file-command-table
 :errorp nil
 :menu '(
	 ("Change Data Directory" :command com-change-data-directory)
	 ("Default Data Directory" :command com-default-data-directory)
	 ("Clear" :command com-clear)
	 ("Quit" :command com-quit)))

(make-command-table
 'sword-command-table
 :errorp nil
 :menu
 '(
   ("Move sword to lexpr" :command com-move-sword-to-lexpr)
   ("Random sword" :command com-random-sword)
   ("Random inf sword" :command com-random-inf-swords)
   ("Join sword lexpr" :command com-join-sword-lexpr)
   ("Alphabet sword" :command com-alphabet-sword)
   ("Malphabet sword" :command com-malphabet-sword)
   ("Size sword" :command com-size-sword)
   ("Parikh vector" :command com-parikh-vector)
   ("Membership to language" :command com-membership-to-language)
   ("Retrieve sword" :command com-retrieve-sword)
   ("Read sword" :command com-read-sword)
   ("Load sword" :command com-load-sword)
   ("Save sword" :command com-save-sword)
   ("Clear sword" :command com-clear-sword)
   ))

(make-command-table
 'spec-command-table
 :errorp nil
 :menu
 '(
   ("Alphabet spec" :command com-alphabet-spec)
   ("Malphabet spec" :command com-malphabet-spec)
   ("Retrieve spec" :command com-retrieve-spec)
   ("Load spec (.txt)" :command com-load-spec)
   ("Save spec" :command com-save-spec)
   ("Clear spec" :command com-clear-spec)
   ))

(make-command-table
 'lexpr-command-table
 :errorp nil
 :menu '(
	 ("Move lexpr to sword" :command com-move-lexpr-to-sword)
	 ("Elementary lexpr?" :command com-elementary-lexpr)
	 ("Language lexpr" :command com-language-lexpr)
	 ("Prefix Language lexpr" :command com-prefix-language-lexpr)
	 ("Cardinal lexpr" :command com-cardinal-lexpr)
	 ("Factorize lexpr" :command com-factorize-lexpr)
	 ("Retrieve lexpr" :command com-retrieve-lexpr)
	 ("Read lexpr" :command com-read-lexpr)
	 ("Load lexpr" :command com-load-lexpr)
	 ("Save lexpr" :command com-save-lexpr)
         ("Clear lexpr" :command com-clear-lexpr)
	 ))

(make-command-table
 'language-command-table
 :errorp nil
 :menu '(
;;	 ("Size language" :command com-size-language)
	 ("Cardinal language" :command com-cardinal-language)
	 ("Alphabet language" :command com-alphabet-language)
	 ("Malphabet language" :command com-malphabet-language)
	 ("Move language to lexpr" :command com-move-language-to-lexpr)
	 ("Rename language" :command com-rename-language)
	 ("Retrieve language" :command com-retrieve-language)
	 ("Read language" :command com-read-language)
	 ("Load language" :command com-load-language)
	 ("Save language" :command com-save-language)
	 ("Clear language" :command com-clear-language)
	 ))

(make-command-table
 'problem-command-table
 :errorp nil
 :menu '(
	 ("Alphabet problem" :command com-alphabet-problem)
	 ("Malphabet problem" :command com-malphabet-problem)
	 ("Add lexpr" :command com-add-lexpr)
	 ("Solve" :command com-solve)))

(make-command-table
 'menubar-command-table
 :errorp nil
 :menu
 '(
   ("File " :menu file-command-table)
   ("Spec " :menu spec-command-table)
   ("Sword " :menu sword-command-table)
   ("Problem " :menu problem-command-table)
   ("Lexpr " :menu lexpr-command-table)
   ("Language " :menu language-command-table)
   )
 )

;; (defmethod display-processname :around ((name string))
;;   (if *processname-pane*
;;       (setf (gadget-value *processname-pane*)
;; 	    (normalize-string name +largeur-texte+))
;;       (format *error-output* "~A" name)))

(defun get-new-directory (stream previous)
  (let ((pathname (accept 'pathname :stream stream
 			  :default previous
  			  :default-type 'pathname
 			  :insert-default t)))
    (if pathname
	(let ((dir (probe-file pathname)))
	  (if dir
	      (namestring dir)
	      (format *output-stream* "~A: no such directory~%" pathname)))
	  (format *output-stream* "Data directory is ~A~%" *data-directory*))
	(format *output-stream* "~A incorrect pathname~%" pathname)))

(defclass data-directory-field (text-field) ())
   
(defgeneric change-data-directory (data-directory-field))

(defmethod change-data-directory ((ddp data-directory-field))
  (let ((dirname (get-new-directory ddp *data-directory*)))
    (when dirname
      (setf *data-directory* dirname))))

;;(defclass data-directory-field (application-pane) ())

(defun display-ddp (frame ddp)
  (declare (ignore frame))
  (format ddp "~A" *data-directory*))

(define-application-frame visu-smots ()
  ()
  (:panes
   (problem-pane
    (make-pane 'clim-stream-pane
	       :name 'problem-pane))
   (lexpr-pane
    (make-pane 'clim-stream-pane
	       :name 'lexpr-pane))
   (language-pane
    (make-pane 'clim-stream-pane
	       :name 'language-pane))
   (sword-pane
    (make-pane 'clim-stream-pane
	       :name 'sword-pane))
   (result-pane
    (make-pane 'clim-stream-pane :name 'result-pane))
   (interactor-pane :interactor)
   (menu-bar
    (climi::make-menu-bar 'menubar-command-table))
;; ;;    (make-pane 'text-field :value *data-directory*
;;     (make-pane 'data-directory-field :height 4 :width :compute
;; 	       :display-function 'display-ddp
;; ;;	       :activate-callback (lambda (x) (change-data-directory x))
;;        ))
   (data-directory-field
    :text-field
    :editable-p nil
    :value *data-directory* :height 4)
;;    (make-pane 'data-directory-field
;;	       :height 4 
;; 	       :armed t
;; 	       :display-function 'display-ddp
;;	       :incremental-redisplay t
;;	       ))
   (quit
    :push-button
    :label "Quit"
    :activate-callback (lambda (x)
			 (declare (ignore x))
			 (com-quit)))
;;    (stop
;;     :push-button
;;     :label "Stop"
;;     :activate-callback (lambda (x)
;; 			 (declare (ignore x))
;; 			   (com-stop)))
   (clear
    :push-button
    :label "clear" ;:max-width 100 :max-height 100
    :activate-callback (lambda (x)
			 (declare (ignore x))
			 (com-clear)))
   (mark-button
    :toggle-button
    :label "Marks"
    :indicator-type :one-of
    :value *show-mark*
    :value-changed-callback
    (lambda (gadget value)
      (declare (ignore gadget value))
      (com-toggle-mark)))
;;    (process-button
;;     :toggle-button
;;     :label "Processes"
;;     :indicator-type :one-of
;;     :value *with-processes*
;;     :value-changed-callback
;;     (lambda (gadget value)
;;       (declare (ignore gadget value))
;;       (com-toggle-processes)))
;;    (processname-pane
;;     :text-field
;;     :editable-p nil
;;     :value (normalize-string "no running process" +largeur-texte+))
;;    (progress-pane
;;     :text-field
;;     :editable-p nil
;;     :value (normalize-string "-" +largeur-texte+))
   )
(:layouts
   (default
       (vertically ()
	menu-bar
	(vertically ()
	  (horizontally (:height 100) interactor-pane)
	  (scrolling (:height 150) problem-pane)
	  (scrolling (:height 150) lexpr-pane)
	  (horizontally (:height 100)
	    (scrolling () sword-pane)
	    (scrolling () language-pane))
	  (scrolling (:height 100) result-pane))
	(horizontally ()
	  quit clear data-directory-field mark-button)))))

(define-condition all-errors (error)
  ())

(defmethod default-frame-top-level :before ((frame application-frame)
     &key (command-parser 'command-line-command-parser)
	  (command-unparser 'command-line-command-unparser)
	  (partial-command-parser 'command-line-read-remaining-arguments-for-partial-command)
	  (prompt "command: "))
  (declare (ignore command-parser command-unparser partial-command-parser prompt))
  (setf
;;   *progress-pane* (climi::find-pane-named *application-frame* 'progress-pane)
   *data-directory-field* (climi::find-pane-named *application-frame* 'data-directory-field)
   *sword-pane* (climi::find-pane-named *application-frame* 'sword-pane)
   *language-pane* (climi::find-pane-named *application-frame* 'language-pane)
   *problem-pane* (climi::find-pane-named *application-frame* 'problem-pane)
   *lexpr-pane* (climi::find-pane-named *application-frame* 'lexpr-pane)
   *output-stream* (climi::find-pane-named *application-frame* 'result-pane)
   )
  (setf *standard-output* *output-stream*)
  )

(defmethod execute-frame-command :around ((frame application-frame) command)
  (call-next-method))

(defun check-problem ()
  (or (problem *spec*)
    (format *output-stream* "No current PROBLEM~%")))

(defun check-language ()
  (or (language *spec*)
    (format *output-stream* "No current Language~%")))

(defun check-sword ()
  (or (sword *spec*)
    (format *output-stream* "No current Sword~%")))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Commands
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
(define-visu-smots-command (com-toggle-mark :name t) ()
  (toggle-show-mark)
  (display-current-spec))

;; (define-visu-smots-command (com-toggle-processes :name t) ()
;;   (toggle-with-processes))

(define-visu-smots-command (com-change-data-directory :name t) ()
   (let ((pathname (accept 'pathname :prompt "data directory?"
 			  :default *data-directory*
 			  :default-type 'pathname
 			  :insert-default t)))
     (if pathname
	 (let ((dir (probe-file pathname)))
	   (if dir 
	       (setf *data-directory* (namestring (probe-file pathname)))
	       (format *output-stream* "no such directory~%")))
	 (format *output-stream* "Data directory is ~A~%" *data-directory*))
    ))

(define-visu-smots-command (com-default-data-directory :name t) ()
  (setf *data-directory* (initial-data-directory))
  (format *output-stream* "Data directory is ~A~%" *data-directory*))

(define-visu-smots-command (com-alphabet-sword :name t) ()
  (let ((sword (sword *spec*)))
    (when sword
      (format *output-stream* "Alphabet of current sword is ~A~%" (alphabet-of sword)))))

(define-visu-smots-command (com-malphabet-sword :name t) ()
  (let ((sword (sword *spec*)))
    (when sword
      (format *output-stream* "Malphabet of current sword is ~A~%" (malphabet-of sword)))))

(define-visu-smots-command (com-malphabet-lexpr :name t) ()
  (let ((lexpr (lexpr *spec*)))
    (when lexpr
      (format *output-stream* "Malphabet of current lexpr is ~A~%" (malphabet-of lexpr)))))

(define-visu-smots-command (com-alphabet-lexpr :name t) ()
  (let ((lexpr (lexpr *spec*)))
    (when lexpr
      (format *output-stream* "Alphabet of current lexpr is ~A~%" (alphabet-of lexpr)))))

(define-visu-smots-command (com-elementary-lexpr :name t) ()
  (let ((lexpr (lexpr *spec*)))
    (when lexpr
      (format
       *output-stream*
       "~:[We cannot say yet whether the current lexpr is~;Current lexpr is~] elementary~%"
       (elementaryp lexpr)))))

(define-visu-smots-command (com-malphabet-language :name t) ()
  (let ((language (language *spec*)))
    (when language
      (format *output-stream* "Malphabet of current language is ~A~%" (malphabet-of language)))))

(define-visu-smots-command (com-malphabet-problem :name t) ()
  (let ((problem (problem *spec*)))
    (when problem
      (format *output-stream*
	      "Malphabet of current problem is ~A~%" (malphabet-of problem)))))

(define-visu-smots-command (com-alphabet-problem :name t) ()
  (let ((problem (problem *spec*)))
    (when problem
      (format *output-stream*
	      "Alphabet of current problem is ~A~%" (alphabet-of problem)))))

(define-visu-smots-command (com-malphabet-spec :name t) ()
  (when *spec*
    (format *output-stream* "Malphabet of current spec is ~A~%" (malphabet-of *spec*))))

(define-visu-smots-command (com-alphabet-spec :name t) ()
  (when *spec*
    (format *output-stream* "Alphabet of current spec is ~A~%" (alphabet-of *spec*))))

(define-visu-smots-command (com-alphabet-language :name t) ()
  (let ((language (language *spec*)))
    (when language
      (format *output-stream* "Alphabet of current language is ~A~%" (alphabet-of language)))))

(define-visu-smots-command (com-size-sword :name t) ()
  (let ((sword (sword *spec*)))
    (when sword
      (format *output-stream* "size of current sword is ~A~%" (size sword)))))

(define-visu-smots-command (com-parikh-vector :name t) ()
  (let ((sword (sword *spec*)))
    (when sword
      (format *output-stream* "The Parikh vector of current sword is ~A~%" (parikh-vector sword)))))

(define-visu-smots-command (com-membership-to-language :name t) ()
  (let ((sword (sword *spec*)))
    (when sword
      (let ((language (language *spec*)))
	(when language
	  (format
	   *output-stream*
	   "Current sword is~:[ not~;~] in current language~%" (member sword (words language))))))))
  
(define-visu-smots-command (com-load-spec :name t) ()
  (let ((name (accept '(data-file-name "*.txt")
		      :prompt "spec filename"
		      :default nil
		      :display-default nil
		      )))
     (when name
       (let ((spec (read-spec-from-path name)))
	 (when spec
;; 	  (com-clear-spec)
 	  (set-and-display-current-spec spec)
 	  )))))

(define-visu-smots-command (com-save-spec :name t) ()
  (let ((*limit-constraints* nil)
	(name (accept 'string :prompt "spec filename"
		      :default "save")))
	  (save-spec (concatenate 'string name ".txt"))))

(define-visu-smots-command (com-save-lexpr :name t) ()
  (let ((name (accept 'string :prompt "lexpr filename"
		      :default "lexpr-save")))
	  (save-lexpr (concatenate 'string name ".txt"))))

(define-visu-smots-command (com-save-language :name t) ()
  (let ((name (accept 'string :prompt "language filename"
		      :default "language-save")))
	  (save-language (concatenate 'string name ".txt"))))

(define-visu-smots-command (com-save-sword :name t) ()
  (let ((name (accept 'string :prompt "sword filename"
		      :default "sword-save")))
	  (save-sword (concatenate 'string name ".txt"))))

(define-visu-smots-command (com-save-spec-readable :name t) ()
  (let ((*limit-constraints* nil)
	(name (accept 'string :prompt "spec filename"
		      :default "save")))
    (save-spec (concatenate 'string name ".txt"))))

(define-visu-smots-command (com-retrieve-spec :name t) ()
  (let ((spec 
	 (accept 'spec
		 :prompt "spec name")))
    (when spec
      (set-and-display-current-spec spec))))

(define-visu-smots-command (com-load-problem :name t) ()
  (let ((name (accept '(data-file-name)
		      :prompt "problem filename"
		      :default nil
		      :display-default nil
		      )))
    (when name
      (let ((constraints (load-file-from-absolute-filename name #'read-constraints)))
	(if constraints
	    (set-and-display-current-problem (make-problem name constraints))
	    (format *output-stream* "no constraints~%"))))))

(define-visu-smots-command (com-read-problem :name t) ()
  (let ((name (accept 'string :prompt "problem name")))
    (when name
      (let* ((str (concatenate 'string (accept 'string :prompt "constraints") " "))
	     (constraints (with-input-from-string (foo str)
			    (read-constraints foo))))
    	(if constraints
	    (set-and-display-current-problem (make-problem name constraints))
	    (format *output-stream* "no constraints~%"))))))

(define-visu-smots-command (com-retrieve-problem :name t) ()
  (let ((problem 
	 (accept 'problem
		 :prompt "problem name"
		 )))
    (when problem
      (set-and-display-current-problem problem))))

(define-visu-smots-command (com-solve :name t) ()
  (let ((problem (problem *spec*)))
    (when problem
      (solve problem)
      (rename-object (result problem) (name problem))
      (clear-language-pane)
      (set-and-display-current-lexpr (result (problem *spec*))))))

;;      (launch-process "Solve" (process-solve problem)))))

(define-visu-smots-command (com-quit :name t) ()
  (frame-exit *application-frame*))

(define-visu-smots-command (com-load-language :name t) ()
  (let ((name (accept 'string
		      :prompt "language filename"
		      :default nil
		      :display-default nil
		      )))
    (when name
      (let ((swords (load-file-from-absolute-filename (absolute-filename name) #'read-swords)))
	(when swords
	  (load-language name (make-language swords)))))))

(define-visu-smots-command (com-read-language :name t) ()
  (let ((name (accept 'string :prompt "name")))
    (when name
      (let ((str (accept 'string :prompt "language")))
	(when str
	    (let ((slanguage (with-input-from-string (foo str)
		       (read-slanguage foo))))
	      (when slanguage
		(load-language name slanguage))))))))

(define-visu-smots-command (com-retrieve-language :name t) ()
   (let ((language 
	 (accept 'language
		 :prompt "language name"
		 )))
    (when language
      (set-and-display-current-language language))))

(define-visu-smots-command (com-clear-language :name t) ()
  (clear-current-language))

(define-visu-smots-command (com-clear-lexpr :name t) ()
  (clear-current-lexpr))

(define-visu-smots-command (com-clear-sword :name t) ()
  (clear-current-sword))

(define-visu-smots-command (com-clear-problem :name t) ()
  (clear-current-problem))

(define-visu-smots-command (com-retrieve-sword :name t) ()
  (let ((sword (accept 'sword :prompt "sword")))
    (when sword
      (set-and-display-current-sword sword))))

(define-visu-smots-command (com-move-sword-to-lexpr :name t) ()
  (let ((sword (sword *spec*)))
    (when sword
      (clear-sword-pane)
      (set-and-display-current-lexpr sword))))

(define-visu-smots-command (com-move-lexpr-to-sword :name t) ()
  (let ((lexpr (lexpr *spec*)))
    (when (and lexpr (swordp lexpr))
      (clear-lexpr-pane)
      (set-and-display-current-sword lexpr))))

(define-visu-smots-command (com-retrieve-lexpr :name t) ()
  (let ((lexpr (accept 'lexpr :prompt "lexpr")))
    (when lexpr
      (set-and-display-current-lexpr lexpr))))

(define-visu-smots-command (com-random-sword :name t) ()
  (let ((sword (random-sword 10 (alphabet-of *spec*))))
    (when sword
      (set-and-display-current-sword sword))))

(define-visu-smots-command (com-random-inf-swords :name t) ()
  (let ((sword (sword *spec*)))
    (when sword
      (let ((newsword (random-inf-sword sword))
	    (lexpr (random-inf-sword sword)))
	(loop
	 do (setf lexpr (random-inf-sword sword))
	 until (joinablep newsword lexpr))
	(when newsword
	  (set-and-display-current-sword newsword))
	(when lexpr
	  (set-and-display-current-lexpr lexpr))))))

(define-visu-smots-command (com-cardinal-lexpr :name t) ()
  (let ((lexpr (lexpr *spec*)))
    (when lexpr
      (format *output-stream* "the cardinal of the language associated with current lexpr is ~A~%" (ccard (lexpr *spec*))))))

(define-visu-smots-command (com-language-lexpr :name t) ()
  (let ((lexpr (lexpr *spec*)))
    (when lexpr
      (let ((language (compute-language lexpr)))
	(set-and-display-current-language language)))))
;;      (launch-process "Language-lexpr" (process-language-lexpr lexpr)))))

(define-visu-smots-command (com-prefix-language-lexpr :name t) ()
  (let ((lexpr (lexpr *spec*)))
    (when lexpr
      (let ((n (accept 'integer :prompt "swords length")))
	(when n
	  (let ((language (compute-prefix-language lexpr n)))
	    (set-and-display-current-language language)))))))

(define-visu-smots-command (com-factorize-lexpr :name t) ()
  (let ((lexpr (lexpr *spec*)))
    (when lexpr
      (let ((newlexpr (factorize lexpr)))
	(set-and-display-current-lexpr newlexpr)))))

(define-visu-smots-command (com-load-sword :name t) ()
  (let ((name (accept '(data-file-name "*.txt")
		      :prompt "sword filename"
		      :default nil
		      :display-default nil)))
    (when name
      (let ((sword (load-file-from-absolute-filename name #'read-sword)))
	(when sword
	  (load-sword sword))))))

(define-visu-smots-command (com-load-lexpr :name t) ()
  (let ((name (accept '(data-file-name "*.txt")
		      :prompt "lexpr filename"
		      :default nil
		      :display-default nil)))
    (when name
      (let ((lexpr (load-file-from-absolute-filename name #'read-lexpr)))
	(when lexpr
	  (load-lexpr lexpr))))))

(define-visu-smots-command (com-join-sword-lexpr :name t) ()
  (let ((sword (sword *spec*)))
    (when sword
      (let ((lexpr (lexpr *spec*)))
	(when lexpr
	  (let ((join (ljoin sword lexpr)))
	    (when join
	      (format *output-stream* "~A~%" join))))))))

(define-visu-smots-command (com-read-sword :name t) ()
  (let ((sword-string (accept 'string
			     :prompt "sword")))
    (when sword-string
      (load-sword
       (with-input-from-string (foo sword-string)
	 (read-sword foo))))))

(define-visu-smots-command (com-read-lexpr :name t) ()
  (let ((lexpr-string (accept 'string
			     :prompt "lexpr")))
    (when lexpr-string
      (load-lexpr
       (with-input-from-string (foo lexpr-string)
	 (read-lexpr foo))))))

(define-visu-smots-command (com-add-lexpr :name t) ()
  (let ((lexpr (lexpr *spec*)))
    (when lexpr
      (add-constraint lexpr (problem *spec*))
      (display-current-problem))))

(define-visu-smots-command (com-rename-language :name t) ()
  (let ((language (language *spec*)))
    (when language
      (let ((name (accept 'string
			  :prompt "name"
			  )))
	(setf (name language) name)
	(display-current-language)))))
     
(define-visu-smots-command (com-size-language :name t) ()
  (let ((language (language *spec*)))
    (when language
      (format *output-stream*
	      "The size of Language ~A is ~A~%" (name language) (size language)))))

(define-visu-smots-command (com-cardinal-language :name t) ()
  (let ((language (language *spec*)))
    (when language
      (format *output-stream*
	      "The cardinal of Language ~A is ~A~%" (name language) (cardinal language)))))

(define-visu-smots-command (com-move-language-to-lexpr :name t) ()
  (let ((language (language *spec*)))
    (when language
      (setf (lexpr *spec*) language)
      (display-current-lexpr))))

(define-visu-smots-command (com-clear :name t) ()
  (window-clear *output-stream*))

;; (define-visu-smots-command (com-stop :name t) ()
;;   (when *thread*
;;     (clim-sys::destroy-process *thread*)
;;     (format *output-stream* "Process stopped~%")
;;     ))

;; (define-visu-smots-command (com-stop :name t) ()
;;   (when (and *with-processes* *thread* (not *stop*))
;;     (clim-sys::with-lock-held (*stop-lock*)
;;       (setf *stop* t))))

(define-visu-smots-command (com-clear-spec :name t) ()
  (set-and-display-current-spec (get-spec "empty")))

;;;;; End of Commands
