(in-package :smots)

(defvar *simplify-concatenation* t)

(defgeneric empty-wordp (o)
  (:documentation ""))

(defgeneric wordtype (o)
  (:documentation ""))

(defgeneric languatype (o)
  (:documentation ""))

(defclass alanguage (named-object alpha-object) 
  ((computed-cardinal :initarg :computed-cardinal
		      :reader cardinal
		      :initform nil)
   (wordtype :reader wordtype :initarg :wordtype :initform 'sword)
   (languagetype :reader languagetype
		 :initarg :languagetype
		 :initform 'slanguage
		 ))
  (:documentation "abstract class for languages"))

(defgeneric cardinal (alanguage)
  (:documentation "gives the number of words in ALANGUAGE (compute when necessary)"))

(defgeneric compute-cardinal (alanguage)
  (:documentation "compute the number of words in ALANGUAGE"))

(defgeneric lintersection (language1 language2)
  (:documentation "intersection of LANGUAGE1 and LANGUAGE2"))

(defgeneric lunion (language1 language2)
  (:documentation "union of LANGUAGE1 and LANGUAGE2"))

(defgeneric lgeneric (language1 language2 fun)
  (:documentation "fun of LANGUAGE1 and LANGUAGE2"))

(defgeneric lgeneric-bool (language1 language2 boolfun)
  (:documentation "boolfun of LANGUAGE1 and LANGUAGE2"))

(defgeneric equivalent (language1 language2)
  (:documentation "comparaison of LANGUAGE1 and LANGUAGE2"))

(defgeneric ljoin (sexpr1 sexpr2)
  (:documentation "ljoin de deux slangages"))

(defgeneric lmirror (language)
  (:documentation "mirror of LANGUAGE"))

(defgeneric priority (alanguage)
  (:documentation "priority of ALANGUAGE"))

(defgeneric factor-extreme-letter (alanguage &optional last)
  (:documentation
   "transform ALANGUAGE into a union of concatenations whose
    first (or last) argument is a possible first (or last) letter"))

(defgeneric factor-first-letter (alanguage)
  (:documentation
   "transform ALANGUAGE into a union of concatenations whose
    first argument is a possible first letter"))

(defgeneric factor-last-letter (alanguage)
  (:documentation
   "transform ALANGUAGE into a union of concatenations whose
    last argument is a possible last letter"))

(defgeneric order-according-priority (alanguages)
  (:documentation "order a list of ALANGUAGE according to the priority of their operators"))

(defgeneric reconstruct (alanguage)
  (:documentation "reconstruct ALANGUAGE according global parameters"))

(defgeneric factorize (alanguage)
  (:documentation "try to factorize ALANGUAGE into a simpler expression"))

(defmethod equivalent ((l1 alanguage) (l2 alanguage))
  (eq l1 l2))

(defmethod equivalent ((languages1 list) (languages2 list))
  (every (lambda (arg1 arg2) (equivalent arg1 arg2))
	 languages1
	 languages2))

(defun alanguagep (o)
  (typep o 'alanguage))

(defmethod priority ((alanguage alanguage))
  6)

(defmethod alphabet-of :around ((alpha-object alpha-object))
;;  (format *error-output* "alphabet-of around ~A ~%" alpha-object)
  (with-slots (alphabet-of) alpha-object
    (or alphabet-of
	(setf alphabet-of (gen-calphabet-of alpha-object 'letter)))))

(defmethod order-according-priority ((args list))
  (sort (copy-list args) #'>
	:key #'(lambda (arg) (priority arg))))

;;; by default, nothing
(defmethod reconstruct ((alanguage alanguage))
  alanguage)

;;; by default, nothing
(defmethod factorize ((alanguage alanguage))
  alanguage)

(defmethod factor ((alanguage alanguage))
  alanguage)

(defgeneric extreme-letters (language &optional last)
  (:documentation 
   "set of the first (resp. last) letters of the words in the LANGUAGE"))

(defgeneric compute-language (alanguage)
  (:documentation "compute the set of words represented by ALANGUAGE"))

(defgeneric compute-prefix-language (alanguage n)
  (:documentation "compute the set of words of length <= n of ALANGUAGE"))

(defgeneric first-letters (alanguage)
  (:documentation "the set of letters that appear at the beginning of words in ALANGUAGE"))

(defgeneric last-letters (alanguage)
  (:documentation "the set of letters that appear at the end of words in ALANGUAGE"))

(defmethod first-letters (o)
  (extreme-letters o))

(defmethod last-letters (o)
  (extreme-letters o 'last))

(defgeneric computed (alanguage)
  (:documentation "T if ALANGUAGE is computed in extension"))

(defgeneric cut-extreme-letter (alanguage letter &optional last)
  (:documentation "residual of ALANGUAGE by LETTER (from end or from beginning"))

(defgeneric cut-first-letter (alanguage letter)
  (:documentation "residual of ALANGUAGE by LETTER (from beginning)"))

(defgeneric cut-last-letter (alanguage letter)
  (:documentation "residual of ALANGUAGE by LETTER (from end)"))

(defmethod cut-first-letter ((alanguage alanguage) (letter abstract-letter))
  (cut-extreme-letter alanguage letter))

(defmethod cut-last-letter ((alanguage alanguage) (letter abstract-letter))
  (cut-extreme-letter alanguage letter 'last))

(defmethod extreme-letters ((alanguages list) &optional (last nil))
;;  (format *error-output* "extreme-letters list~%")
  (remove-duplicates
   (mappend #'(lambda (alanguage) (extreme-letters alanguage last))
	    alanguages)))

(defgeneric concatenation (language1 language2)
  (:documentation 
   "concatenation of the 2 languages preserving existing marks"))

(defgeneric concatenation-n (language1 language2)
  (:documentation 
   "concatenation of the 2 languages with numbering"))

(defgeneric ccard (alanguage)
  (:documentation 
   "cardinal of ALANGUAGE; compute ALANGUAGE if not already computed"))

(defmethod ccard ((alanguage alanguage))
  (cardinal (compute-language alanguage)))

(defgeneric residual (alanguage letters)
  (:documentation "residual of a language by a word or a sequence of letters"))

(defgeneric contains-empty-sword (alanguage)
  (:documentation "t if ALANGUAGE contains [])"))

(defgeneric reduce-lunion-on-prefix-languages (fsls sexpr n)
  (:documentation ""))

(defmethod residual ((alanguage alanguage) (letters list))
  (if (endp letters)
      alanguage
      (residual (cut-first-letter alanguage (car letters)) (cdr letters))))

(defmethod elementaryp ((alanguage alanguage))
  (declare (ignore alanguage))
  nil)

;; by default we compute the whole language
;; we must try to do better when possible
(defmethod compute-prefix-language ((lexpr alanguage) (n integer))
;;  (format *error-output* "compute-prefix-language op-lexpr~%")
  (let ((l (compute-language lexpr)))
    (when l
      (compute-prefix-language l n))))

(defgeneric gen-alphabet-of (object lettertype)
  (:documentation "returns the alphabet of the letters present in OBJECT"))

(defmethod gen-alphabet-of (o (lettertype (eql 'letter)))
  (alphabet-of o))

(defgeneric gen-calphabet-of (object lettertype)
  (:documentation "returns the alphabet of the letters present in OBJECT"))

(defmethod gen-alphabet-of ((l list) lettertype)
;;  (format *error-output* "gen-alphabet-of list ~A ~%" lettertype)
  (gen-calphabet-of l lettertype))

(defmethod gen-calphabet-of ((l list) lettertype)
;;  (format *error-output* "gen-calphabet-of list ~A ~%" lettertype)
  (make-alphabet
   (mappend (lambda (e) (letters (gen-alphabet-of e lettertype)))
	    l)
   lettertype))

(defmethod alphabet-of (o)
;;  (format *error-output* "alphabet-of t ~A ~%"o)
  (gen-calphabet-of o 'letter))

(defmethod gen-calphabet-of :around ((alpha-object alpha-object) lettertype)
;;  (format *error-output* "gen-alphabet-of around alpha-object ~A ~A~%" (type-of alpha-object) lettertype)
  (if (eq lettertype (type-of alpha-object))
      (make-alphabet (list alpha-object) lettertype)
      (call-next-method)))
