(in-package :smots)

(defgeneric letter-number (letter)
  (:documentation "internal number of LETTER, useful for sorting letters"))

(defgeneric letter-counter (specification)
  (:documentation "value of internal counter of letters in specification"))

(defgeneric (setf letter-counter) (val specification)
  (:documentation "setf of internal counter of letters in specification"))

(defgeneric letters (object)
  (:documentation "list of all the letters used in OBJECT"))

(defgeneric (setf letters) (val spec)
  (:documentation "writer on the list of letters defined in SPECIFICATION"))

(defvar *internal-letters* nil)

(defmethod letters ((spec (eql nil)))
  *internal-letters*)

(defmethod (setf letters) (val (spec (eql nil)))
  (setf *internal-letters* val))

(defgeneric alphabet-of (alpha-object)
  (:documentation
   "alphabet of the letters present in ALPHA-OBJECT"))

(defclass alpha-object ()
   ((alphabet-of :accessor alphabet-of
		 :initform nil )))
  
(defclass abstract-letter (alpha-object)
  ((number :initform (incf (letter-counter *spec*)) :reader letter-number))
  (:documentation "class for general letters"))

(defclass letter (abstract-letter named-object) ()
 (:documentation "class for general letters"))

(defmethod lettertype ((abstract-letter abstract-letter))
  (type-of abstract-letter))

(defmethod print-object ((letter letter) stream)
  (write-name (name letter) stream))

(defgeneric make-letter (string)
  (:documentation "created a letter STRING"))

(defmethod make-letter ((string string))
  (let ((name (make-name string)))
    (or (find-object name (letters *spec*)
		     :test #'(lambda (name letter)
			       (eq name (name letter))))
	(let ((letter (make-instance 'letter :name name)))
	  (push letter (letters *spec*))
	  letter))))

(defgeneric make-unique-internal-letter ())

(defmethod make-unique-internal-letter ()
  (make-instance 'letter :name (symbol-name (gensym))))

(defgeneric letter-value (mletter)
  (:documentation "an artificial value for a letter for sorting letters"))

(defmethod letter-value ((letter abstract-letter))
  (letter-number letter))

(defun sort-letters (letters)
  (sort (copy-list letters) #'< :key #'letter-value))

(defun sortedp (letters)
  (let ((copy (copy-list letters)))
    (equal (sort-letters copy) letters)))

(defgeneric equiv-named-letters (letter1 letter2)
  (:documentation "T if LETTER1 and LETTER2 are equivalent"))

(defmethod equiv-named-letters ((letter1 abstract-letter) (letter2 abstract-letter))
  (eq (name letter1) (name letter2)))

(defmethod gen-calphabet-of (o (lettertype (eql 'letter)))
  (format *error-output* "gen-calphabet-of o ~A letter bidon on ne devrait pas passer par là ~%" (type-of o)))
