(in-package :smots)

(defvar *show-mark* nil)

(defun toggle-show-mark ()
  (setf *show-mark* (not *show-mark*)))

(defgeneric mletters (specification)
  (:documentation "list of all the mletters defined in SPECIFICATION"))

(defgeneric (setf mletters) (val specification)
  (:documentation "writer on the list of mletters defined in SPECIFICATION"))

(defvar *internal-mletters* nil)

(defmethod mletters ((spec (eql nil)))
  *internal-mletters*)

(defmethod (setf mletters) (val (spec (eql nil)))

  (setf *internal-mletters* val))

(defgeneric malphabet-of (object)
  (:documentation "alphabet of the mletters present in OBJECT"))

(defclass m-mixin ()
  ((malphabet-of :accessor malphabet-of :initform nil))
  (:documentation "for languages with marked letters"))

(defclass mletter (abstract-letter m-mixin)
  ((letter :initarg :letter :reader letter)
   (mark :initarg :mark :reader mark))
  (:documentation "class for markd letters"))

(defmethod malphabet-of :around ((m-mixin m-mixin))
;;  (format *error-output* "malphabet-of around m-mixin ~A ~%" m-mixin)
  (with-slots (malphabet-of) m-mixin
;;    (format *error-output* "slot malphabet ~A ~%" malphabet-of)
    (or malphabet-of
	(setf malphabet-of (gen-calphabet-of m-mixin 'mletter)))))

(defun mletterp (o)
  (typep o 'mletter))

(defgeneric make-mletter (string &optional mark)
  (:documentation "return an mletter with mark MARK from the letter of name STRING"))

(defmethod make-mletter ((string string) &optional (mark 0))
  (let* ((letter (make-letter string))
	 (name (name letter)))
    (or (find-object name (mletters *spec*)
		     :test (lambda (name mletter)
			       (and (eq name (name mletter))
				    (= mark (mark mletter)))))
      (let ((mletter (make-instance 'mletter
				    :letter letter
				    :mark mark)))
	(setf (mletters *spec*) (append (mletters *spec*) (list mletter)))
	mletter))))
 
(defmethod name ((mletter mletter))
  (name (letter mletter)))

(defmethod letter-number ((mletter mletter))
  (letter-number (letter mletter)))

(defmethod make-mletter ((letter letter) &optional (mark 0))
  (make-mletter (name letter) mark))

(defmethod print-object ((mletter mletter) stream)
  (format stream "~A" (letter mletter))
  (let ((mark (mark mletter)))
    (when ;;(and
	*show-mark* ;; (plusp mark))
      (format stream "_~D" mark))))

(defmethod letter-value ((mletter mletter))
  (let ((n (letter-number mletter))
	(ind (mark mletter)))
    (+ (* 10 n) ind)))

(defun letters-from-mletters (mletters)
  (remove-duplicates (mapcar #'letter mletters)
		     :test #'eq))

(defun make-malphabet (mletters)
  (make-alphabet mletters 'mletter))

(defun alphabet-from-malphabet (malphabet)
  (make-alphabet
   (mapcar #'letter
	   (letters malphabet))
   'letter))

(defgeneric malphabet-of (alpha-object))

(defmethod malphabet-of (o)
;;  (format *error-output* "malphabet ~A~%" o)
  (gen-calphabet-of o 'mletter))

(defmethod gen-alphabet-of (o (lettertype (eql 'mletter)))
  (malphabet-of o))

(defmethod gen-calphabet-of ((mletter mletter) lettertype)
;;  (format *error-output* "gen-calphabet-of mletter t ~A~%" mletter)
  (gen-alphabet-of (letter mletter) lettertype))
