(in-package :smots)

(defvar *simplify-sletter-output* nil)

(defgeneric sletters (specification)
  (:documentation "current sletters of SPECIFICATION"))

(defgeneric (setf sletters) (val specification)
 (:documentation "writer on current sletters of SPECIFICATION"))

(defvar *internal-sletters* nil)

(defmethod sletters ((spec (eql nil)))
  *internal-sletters*)

(defmethod (setf sletters) (val (spec (eql nil)))
  (setf *internal-sletters* val))

(defclass s-mixin (m-mixin)
  ((slalphabet-of :accessor slalphabet-of :initform nil))
  (:documentation "for languages whose words are Swords"))

(defclass sletter (word abstract-letter s-mixin) ())

(defmethod slalphabet-of :around ((s-mixin s-mixin))
  (with-slots (slalphabet-of) s-mixin
    (or slalphabet-of
	(setf slalphabet-of (gen-calphabet-of s-mixin 'sletter)))))

(defun sletterp (o)
  (typep o 'sletter))

(defun slettersp (l)
  (every #'sletterp l))

(defmethod print-object ((sletter sletter) stream)
  (let ((mletters (letters sletter)))
    (if (and *simplify-sletter-output* (null (cdr mletters)))
	(prin1 (car mletters) stream)
	(progn 
	  (write-char #\{ stream)
	  (display-sequence mletters stream :sep ",")
	  (write-char #\} stream)))))

(defgeneric no-duplicated-letter (mletters)
  (:documentation
   "true if no two mletters in MLETTERS have the same corresponding
    letter"))

(defmethod no-duplicated-letter ((mletters list))
  (let ((letters (letters-from-mletters mletters)))
    (= (length mletters)
       (length letters))))

(defgeneric make-sletter (mletters)
  (:documentation "return the sletter whose mletters are MLETTERS"))

(defmethod make-sletter ((mletters list))
  (assert (every (lambda (letter) (typep letter 'mletter)) mletters))
  (assert (no-duplicated-letter mletters))
  (assert (= (length mletters)
	     (length (remove-duplicates mletters
					:test #'=
					:key #'letter-number))))
  (setf mletters (sort-letters mletters))
  (or (find-object
       mletters (sletters *spec*)
       :test (lambda (mletters sletter)
	       (equal mletters (letters sletter))))
      (let ((sletter (make-instance 'sletter :letters mletters
					     :lettertype 'mletter)))
	(push sletter (sletters *spec*))
	sletter)))

(defgeneric empty-sletterp (sletter)
  (:documentation "true if SLETTER is the empty one"))

(defmethod empty-sletterp ((sletter sletter))
  (endp (letters sletter)))

(defun make-slalphabet (sletters)
  (make-alphabet sletters 'sletter))

(defgeneric arity (sletter)
  (:documentation "number of mletters in SLETTER"))

(defmethod arity ((sletter sletter))
  (length (letters sletter)))

(defgeneric projection-mletters (mletters alphabet))

(defmethod projection-mletters ((mletters list) (alphabet alphabet))
  (intersection
   mletters
   (letters alphabet)
   :test  (if (eq (type-of (car (letters alphabet))) 'letter)
	      #'equiv-named-letters
	      #'eq)))

(defmethod projection ((sletter sletter) (alphabet alphabet))
  (make-sletter
   (projection-mletters
    (letters sletter)
    alphabet)))

(defun malphabet-from-slalphabet (slalphabet)
  (make-alphabet
   (mappend #'letters
	    (letters slalphabet))
   'mletter))

(defmethod slalphabet-of (o)
  (gen-calphabet-of o 'sletter))

(defmethod gen-alphabet-of (o (lettertype (eql 'sletter)))
  (slalphabet-of o))

(defmethod gen-calphabet-of ((sletter sletter) lettertype)
;;  (format *error-output* "gen-calphabet-of mletter t ~A~%" mletter)
  (gen-alphabet-of (letters sletter) lettertype))
